@@ -2,8 +2,19 @@ import matplotlib.pyplot as plt
 from .stockindc import *
 from .gen import *
 
-# 計算 YOY
+abbr = ['公司代號', '公司名稱', '毛利率(%)', '營業利益率(%)', '稅後純益率(%)', '資產報酬率(%)', 
+# '負債比率(%)', '現金流量比率(%)', '流動比率(%)', 
+ '應收款項週轉率(次)', '每股盈餘(元)', '最新價格', '本益比', '年報公佈後收盤價', '相對 PE1 比率', 
+ '營業收入 YOY(%)' ,'近3月營收 平均YOY(%)', '毛利率 YOY(%)', '營業利益率 YOY(%)', '稅後純益率 YOY(%)', '資產報酬率 YOY(%)'
+]
+
+def lessColumns(d):
+    available = [i for i in abbr if i in d.columns]
+    return d[available]
+    
+# 計算 YOY, 需調用當年度及去年度資料
 def add_indicator_growth(d_current, d_last, special_indicator=None):
+    # d_current 取 FinanceReport, d_last 取 Ratio
     cmpr = ['營業收入(百萬元)', '毛利率(%)', '營業利益率(%)', \
            '稅後純益率(%)', '資產報酬率(%)', '權益報酬率(%)', '現金再投資比率(%)'] if special_indicator==None else special_indicator
 
@@ -23,11 +34,18 @@ def add_indicator_growth(d_current, d_last, special_indicator=None):
     
     # 合併 今年資料 與 算出之成長百分比
     d_res = pd.merge(d_current, d_cmpr, how='inner', on=['公司代號'])
-    d_res = d_res.fillna(1e-4)
-    
+    d_res = d_res.fillna(1e-4)    
     return d_res
-
-# 還原價格
+    
+# 調用 add_indicator_growth 以產出含年成長率之衍生性報表
+def make_drv_rpt(y):
+    rpt = pd.read_csv(f'./FinanceReport/FinanceReport_{y}.csv')
+    fnt = pd.read_csv(f'./Ratio/Ratio_{y-1}.csv')
+    drv_rpt = add_indicator_growth(rpt, fnt)
+    drv_rpt.公司代號 = drv_rpt.公司代號.astype(str)
+    return drv_rpt    
+    
+# 調用股利及減資以推算還原價格
 class AdjPrice:    
     def __init__(self,p_current,p_next):
         self.p_current=p_current
@@ -173,6 +191,7 @@ class AdjPrice:
                 data = recalculate_rdc(i,data)
         self.data = data
 
+# 調用 AdjPrice 產出還原之價格
 def YearlyPrice(yr):
     p_1 = allPrice(FirstTxnDate(yr))
     p_2 = allPrice(FirstTxnDate(yr+1, mh=4, dt=10, reverse=True))
@@ -182,6 +201,7 @@ def YearlyPrice(yr):
     x.adjRdc()
     return x.data
 
+# 預設與 0050 比較
 def YearlyComparingStock(yr,comparing_stock = '0050'):
     tgt_1 = siiPrice(FirstTxnDate(yr), corporation_only=False)
     tgt_2 = siiPrice(FirstTxnDate(yr+1,reverse=True), corporation_only=False)
@@ -192,7 +212,7 @@ def YearlyComparingStock(yr,comparing_stock = '0050'):
     x.adjEtf()
     return x.data
 
-# 還原價格 - 年報酬 - 同期台灣50之報酬
+# 合併多資料以計算回測之報酬 (還原價格 & 年報酬 & 同期台灣50之報酬)
 class BackTesting:    
     def __init__(self,financial_data,adjusted_price,comparing_stock):
         self.financial_data=financial_data
@@ -221,11 +241,34 @@ class BackTesting:
                                 if i!='N']) / len(list(self.data['勝過台灣50'])) * 100, 2)
 
     def introduction(self):
-        print(f'use {self.statement_year} finance statement (which cannot be accessed till {int(self.start_ymd)-1}) to evaluate performance from {self.start_ymd} to {self.end_ymd}')
-        
+        print(f'use {self.statement_year} finance statement (which cannot be accessed till \
+{int(self.start_ymd)-1}) to evaluate performance from {self.start_ymd} to {self.end_ymd}')
+    
+    
+    
+                
+    
+    
     @staticmethod
-    def import_criteria(testing_object,function_with_criteria,verbose=True):
-        data = function_with_criteria(testing_object.data)                    
+    def parse_criteria(df, crts):
+        for i in crts:
+            df = df[i]
+        return df
+
+    # 基底為待回測對象，讀取設定財務指標的條件(function_with_criteria)，計算報酬
+    @staticmethod
+    def import_criteria(testing_object,function_with_criteria,verbose=True):            
+        ################# example function_with_criteria #################
+        # def example_function(d):
+        #     c = [
+        #          d['資產報酬率 YOY(%)']>2.5
+        #         ,d['本益比']<35
+        #         ,d['近3月累計營收 YOY(%)']>-2.5
+        #         ]
+        #     return drv.BackTesting.parse_criteria(d, c)
+        ################# example function_with_criteria #################
+        data = function_with_criteria(testing_object.data)        
+
         rtn = np.array(data['年報酬(%)'])
         rtn = rtn[~np.isnan(rtn)]
         rtn = np.array(list(filter(lambda x:x!=np.inf, rtn)))
@@ -237,100 +280,41 @@ class BackTesting:
             print(f'台灣50指數 {testing_object.comparing_metrics[0]} -> {testing_object.comparing_metrics[1]}, 年報酬 = {testing_object.comparing_metrics[2]}%, 未篩選前隨機個股報酬>台灣50比例 = {testing_object.win_rate}%')
             print(f'篩選股數 = {data.shape[0]}, 年報酬 = {yearly_rtn}%, 超額報酬 = {excess_rtn}%, 篩選個股報酬>台灣50比例 = {win_rate}%\n')
         return data, yearly_rtn
-                
-    @staticmethod
-    def parse_criteria(df, crts):
-        for i in crts:
-            df = df[i]
-        return df
-    
-################# example function_with_criteria #################
-# def example_function(d):
-#     c = [
-#          d['資產報酬率 YOY(%)']>2.5
-#         ,d['本益比']<35
-#         ,d['近3月累計營收 YOY(%)']>-2.5
-#         ]
-#     return drv.BackTesting.parse_criteria(d, c)
-################# example function_with_criteria #################
 
+# 產出指定年度回測的資料
+def GenerateBackTesting(y, verbose=True):
+    # 衍生性報表
+    drv_rpt = make_drv_rpt(y)
 
-
-
-# 第一次需要執行
-def download(x, start_year, end_year,force_writing=False):
-    def check_then_write(function, dir_name):
-        # check if dir exist
-        if not os.path.exists(f'./{dir_name}'):
-            os.mkdir(f'./{dir_name}')
-            
-        for y in range(start_year,end_year+1):
-            # check if file exist
-            file_path = os.path.join(f'./{dir_name}', f'{dir_name}_{y}.csv')
-            print(f'current addressing file = {file_path}')
-            if os.path.exists(file_path):
-                if force_writing==True:
-                    pass
-                else:
-                    writing_check = input('override existing file? ')
-                    if writing_check!='y':
-                        break
-                if function in (YearlyRatio,YearlyPrice,YearlyComparingStock): 
-                    data = function(y)
-                elif function==YearlyFinanceReport: 
-                    data = function(y, import_ratio=True) 
-                else: 
-                    data = function(y, adjusting_report=dir_name)
-                data.to_csv(file_path, index=False, encoding='utf_8_sig')
-            else:
-                if function in (YearlyRatio,YearlyPrice,YearlyComparingStock): 
-                    data = function(y)
-                elif function==YearlyFinanceReport: 
-                    data = function(y, import_ratio=True) 
-                else: 
-                    data = function(y, adjusting_report=dir_name)
-                data.to_csv(file_path, index=False, encoding='utf_8_sig')
-
-    if x=='ratio':
-        check_then_write(YearlyRatio, 'fnt')
-    elif x=='rpt':
-        check_then_write(YearlyFinanceReport, x)
-    elif x in ('etf', 'div', 'rdc'):
-        check_then_write(YearlyAdj, x)
-    elif x=='adjusting_price':
-        check_then_write(YearlyPrice, x)
-    elif x=='comparing_stock':
-        check_then_write(YearlyComparingStock, x)
-
-def SingleYearTesting(y, criteria, verbose=True):
-    rpt = pd.read_csv(f'./rpt/rpt_{y}.csv')
-        
-    fnt = pd.read_csv(f'./fnt/fnt_{y-1}.csv')
-    
-    drv_rpt = add_indicator_growth(rpt, fnt)
-    drv_rpt.公司代號 = drv_rpt.公司代號.astype(str)
-    
-    ajst_price = pd.read_csv(f'./adjusting_price/adjusting_price_{y+1}.csv')
+    # 還原價格
+    ajst_price = pd.read_csv(f'./AdjustingPrice/AdjustingPrice_{y+1}.csv')
     ajst_price.公司代號 = ajst_price.公司代號.astype(str)
     for stock_date in ('股價日期_x', '股價日期_y'):
         ajst_price[stock_date] = ajst_price[stock_date]\
                                     .apply(lambda x:datetime.strptime(x, "%Y-%m-%d") if type(x)==str else x)
-        
-    cmpr_stock = pd.read_csv(f'./comparing_stock/comparing_stock_{y+1}.csv')
+    # 與台灣 50 作比較
+    cmpr_stock = pd.read_csv(f'./ComparingStock/ComparingStock_{y+1}.csv')
     cmpr_stock.公司代號 = cmpr_stock.公司代號.astype(str)
     for stock_date in ('股價日期_x', '股價日期_y'):
         cmpr_stock[stock_date] = cmpr_stock[stock_date]\
                                     .apply(lambda x:datetime.strptime(x, "%Y-%m-%d") if type(x)==str else x)
 
+    # 倒入前述資料以計算出該年度回測
     test_object = BackTesting(drv_rpt,ajst_price,cmpr_stock)
     if verbose==True:
         test_object.introduction()
     test_object.getRvn()
-    test_object.get_comparing_metrics()
+    test_object.get_comparing_metrics()    
+    return test_object
+
+# 單一年度回測
+def SingleYearTesting(y, criteria, verbose=True):
+    test_object = GenerateBackTesting(y, verbose=verbose)
     data, rtn_for_stock = BackTesting.import_criteria(test_object, criteria, verbose=verbose)
     return test_object, data, rtn_for_stock
 
-def SearchBestCriteria(start_year, end_year, criteria, verbose=True, graph=True):
+# 多年度之回測並繪出圖表
+def MultipleYearTesting(start_year, end_year, criteria, verbose=True, graph=True):
     prm_list = []
     statement_time_period = list(range(start_year, end_year+1))
     for y in range(start_year, end_year+1):
@@ -361,4 +345,108 @@ def SearchBestCriteria(start_year, end_year, criteria, verbose=True, graph=True)
         if len(price_time_period) not in (3,5,8):
             print(f'近 {len(rtn_for_tw50_array)} 年報酬 tw50={hstRevn(rtn_for_tw50_array)}% v.s 自選etf={hstRevn(rtn_for_criteria_array)}%')
     d_criteria = dict(zip([i-1 for i in price_time_period], data_of_criteria_array))
-    return d_criteria, rtn_for_tw50_array, rtn_for_criteria_array
\ No newline at end of file
+    return d_criteria, rtn_for_tw50_array, rtn_for_criteria_array
+
+# 一次下載多年資料
+def download(x, start_year, end_year, force_writing=False):
+    def check_then_write(function, dir_name):
+        # check if dir exist
+        if not os.path.exists(f'./{dir_name}'):
+            os.mkdir(f'./{dir_name}')
+        print("data range =", list(range(start_year,end_year+1)))
+        for y in range(start_year,end_year+1):
+            # check if file exist
+            file_path = os.path.join(f'./{dir_name}', f'{dir_name}_{y}.csv')
+            print(f'current addressing file = {file_path}')
+            if os.path.exists(file_path):
+                if force_writing==True:
+                    pass
+                else:
+                    writing_check = input('override existing file? ')
+                    if writing_check!='y':
+                        break
+                if function in (YearlyRatio,YearlyPrice,YearlyComparingStock): 
+                    data = function(y)
+                elif function==YearlyFinanceReport: 
+                    data = function(y, import_ratio=True) 
+                else: 
+                    data = function(y, adjusting_report=dir_name)
+                data.to_csv(file_path, index=False, encoding='utf_8_sig')
+            else:
+                if function in (YearlyRatio,YearlyPrice,YearlyComparingStock): 
+                    data = function(y)
+                elif function==YearlyFinanceReport: 
+                    data = function(y, import_ratio=True) 
+                else: 
+                    data = function(y, adjusting_report=dir_name)
+                data.to_csv(file_path, index=False, encoding='utf_8_sig')
+
+    if x=='Ratio':
+        check_then_write(YearlyRatio, x)
+    elif x=='FinanceReport':
+        check_then_write(YearlyFinanceReport, x)
+    elif x in ('Etf', 'Divident', 'CptlReduct'):
+        check_then_write(YearlyAdj, x)
+    elif x=='AdjustingPrice':
+        check_then_write(YearlyPrice, x)
+    elif x=='ComparingStock':
+        check_then_write(YearlyComparingStock, x)
+
+# 取得呼叫之時點近三個月的營收
+def latest_rvnu():
+    # get current month
+    now = datetime.now()
+    now_last1_mth = add_months(now.date(), -1).month
+    now_last2_mth = add_months(now.date(), -2).month
+    latest_mth = now_last1_mth if now.day>=8 else now_last2_mth
+    return last3MthRvnuGrw(now.year, last_m=latest_mth) 
+
+# 將 年報公佈後收盤價 改成當天的價格 並將本益比做調整
+def update_drv_rpt(drv_rpt):
+    tdy = date.today()
+    
+    print(f"current date = {tdy.strftime('%Y-%m-%d')}")
+    cols = drv_rpt.columns
+    cols = ['最新價格' if x=='年報公佈後收盤價' else x for x in cols]
+    
+    # 近期營收
+    try:
+        lr = latest_rvnu() 
+    except:
+        print("latest_rvnu fails")
+        return drv_rpt
+        
+    # price - 最新價格
+    try:
+        dt = FirstTxnDate(eg_yr=tdy.year, mkt='sii', mh=tdy.month, dt=tdy.day)
+        price = allPrice(dt)
+        price = price.drop('股價日期', axis=1).rename(columns = {'收盤價':'最新價格'})
+    except:
+        print("price fails")
+        return drv_rpt
+        
+    # merge & process data  
+    drv_rpt = drv_rpt.drop(['近3月營收 平均YOY(%)','近3月累計營收 YOY(%)', '年報公佈後收盤價', 'PE1 中位數', 'PE2 中位數'], axis=1)   
+    for i in [drv_rpt,lr,price]:
+        i['公司代號'] = i['公司代號'].astype(str)
+    d = reduce(lambda x,y:pd.merge(x,y,on=['公司代號','公司名稱'],how='left'), [drv_rpt,lr,price])
+    
+    # calculate PE again
+    def flt(x):
+        try:
+            return float(x)
+        except ValueError:
+            return 1e-4
+    for i in ['近3月營收 平均YOY(%)','最新價格','近3月累計營收 YOY(%)']:
+        d[i] = d[i].apply(flt)  
+        
+    d['本益比'] = round(d['最新價格'] / d['每股盈餘(元)'], 2)
+    pe1 = d.groupby('主產業')['本益比'].median().reset_index().rename(columns={'本益比':'PE1 中位數'})
+    pe2 = d.groupby('次產業')['本益比'].median().reset_index().rename(columns={'本益比':'PE2 中位數'})
+    d = pd.merge(pd.merge(d,pe1),pe2)
+    for i in ['PE1 中位數', 'PE2 中位數']:
+        d[i] = d[i].replace(0, 1e-4)
+    d['相對 PE1 比率'] = round(d['本益比'] / d['PE1 中位數'], 4)
+    d['相對 PE2 比率'] = round(d['本益比'] / d['PE2 中位數'], 4)
+    d = d[cols]    
+    return d
\ No newline at end of file
