@@ -11,6 +11,7 @@ from functools import reduce
 import json
 import os
 from collections import Counter
+import random
 
 # functions for processing convenience
 def revr(data, nmbr):
@@ -66,6 +67,7 @@ def siiPrice(str_date, corporation_only=True):
         df = df[~df.證券代號.str.startswith('0')] 
         df = df[df.證券代號.apply(lambda x:len(x)==4)]
     df = clean(df)
+    time.sleep(30+random.randint(0, 10))
     return df
 
 def otcPrice(str_date, corporation_only=True):
@@ -89,6 +91,7 @@ def otcPrice(str_date, corporation_only=True):
         df = df[~df.代號.str.startswith('0')] 
         df = df[df.代號.apply(lambda x:len(x)==4)]
     df = clean(df)
+    time.sleep(30+random.randint(0, 10))
     return df
 
 def Price(dt, mkt):
@@ -158,6 +161,7 @@ class FinanceRatio:
         r.encoding = 'utf8'
         dfs = pd.read_html(StringIO(r.text))
         self.raw_data = dfs
+        time.sleep(30+random.randint(0, 10))
         
     def unify(self):
         if self.purpose=='營益分析':
@@ -216,6 +220,7 @@ class Revenue:
         r.encoding = 'big5'
         dfs = pd.read_html(StringIO(r.text))
         self.raw_data = dfs
+        time.sleep(30+random.randint(0, 10))
         
     def unify(self, smp=True):
         data = pd.concat([data for data in self.raw_data if data.shape[1] <= 11 and data.shape[1] > 5]\
@@ -303,6 +308,7 @@ class EtfDiv:
         data = pd.DataFrame(data=d['data'], columns=d['fields'])
         data = data[data.股票代號.str.startswith('0')]
         self.raw_data = data
+        time.sleep(30+random.randint(0, 10))
 
     def unify(self):
         self.raw_data['詳細資料'] = self.raw_data['詳細資料'].apply(lambda x:x.split("'")[1][9:])
@@ -318,7 +324,7 @@ class EtfDiv:
             r = requests.get(url, headers=headers)
             d = pd.read_html(StringIO(r.text))[0]
             tgt = d.iloc[4][1].split(' ')[0]
-            time.sleep(3)
+            time.sleep(30+random.randint(0, 10))
             return tgt
 
         def cashDividend(dtl):
@@ -328,7 +334,7 @@ class EtfDiv:
             r = requests.get(url, headers=headers)
             d = pd.read_html(StringIO(r.text))[0]
             tgt = d.iloc[2][1].split(' ')[0]
-            time.sleep(3)
+            time.sleep(30+random.randint(0, 10))
             return tgt
         
         x1 = self.raw_data[self.raw_data['權/息']=='息']
@@ -417,6 +423,8 @@ class CptlReduct:
             data = pd.DataFrame(data=d['data'], columns=d['fields'])
             data = data.rename(columns={'停止買賣前收盤價格':'停止買賣前價格'})
             self.raw_data = data
+            time.sleep(30+random.randint(0, 10))
+    
         elif self.mkt_type=='otc':
             r = requests.get(self.url, headers=self.headers)  
             d = pd.read_csv(StringIO(r.text), header = ["股票代號" in l for l in r.text.split("\n")].index(True))
